#ARFF - Logs to ARFF files, used by Weka (http://www.cs.waikato.ac.nz/ml/weka/)
# and converts to .CSV
#Copyright (C) 2005  Christopher J. Hazard
# ncsu at cjhazard dot edu swap the first two

#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.

#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.

#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package ARFF;
require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(ARFFToCSV InitializeARFF ARFFParam ARFFResult ARFFError ARFFFinish);
use strict;
use File::Slurp qw(read_file write_file);

our $ARFF; #filehandle to output

#the number of results, kept for writing out errors
our $num_result_entries;

##############
#ARFFToCSV
# first parameter is the filename of a .arff, second is the name of a .csv
# reads in the file specified by the first parameter and writes it out to the second
sub ARFFToCSV($$) {
  my ($arff_name, $csv_name) = @_;

  #read all meaningful lines (not relation, data indicators, or blanks)
  my @lines = grep !/^\@relation|^$|^\@data/, read_file($arff_name);

  #pull off the data that is not attributes
  my @data = grep !/\@attribute/, @lines;
  #pull off the attributes, replacing each with just its name
  my @attrs = grep s/\@attribute\s+'([^']*)'.*?\n/$1/, @lines;

  write_file($csv_name, join(',', @attrs), "\n", @data);
}

#opens the ARFF file and writes out the header
# pass in the relation name, filename,
#  a pointer to the in-order parameter array,
#  and a reference to an array of any result names
sub InitializeARFF($$$$) {
  my ($filename, $relation_name, $param_list, $result_names) = @_;
  $param_list ||= [];
  #write ARFF file header
  open $ARFF, ">$filename";
  print $ARFF "\@relation $relation_name\n\n";

  #write out attribute list
  foreach my $p (@$param_list) {
    my ($param_name, $vals) = ($p->{name}, $p->{allowed});
    $param_name =~ s/^\$//; #get rid of initial $
    if($vals->[0] =~ /^[a-zA-Z]/) {
      #put in the format {'A','B','C'}
      my $type_list = join ',', (map {"'$_'" } @$vals);
      print $ARFF "\@attribute '$param_name' \{$type_list\}\n";
    }
    else {
      print $ARFF "\@attribute '$param_name' real\n";
    }
  }

  #write out results' attributes
  print $ARFF "\@attribute '$_' real\n" foreach @$result_names;

  $num_result_entries = scalar @$result_names; #keep for writing out errors

  print $ARFF "\n\@data\n\n";
} #InitializeARFF

#writes a list of parameters
sub ARFFParam(@) {  print $ARFF map {$_, ','} @_ }

#writes a list of results
sub ARFFResult(@){  print $ARFF join(',', @_), "\n" }

#writes out the results as errors
sub ARFFError() { ARFFResult split '', '?'x$num_result_entries }

sub ARFFFinish() { close $ARFF }

1;

