#ParameterTuning - Allows easy use of Design of Experiment and
# Recursive Random Search techniques for automating simulations and models.
#Copyright (C) 2005  Christopher J. Hazard
# ncsu at cjhazard dot edu swap the first two

#This library is free software; you can redistribute it and/or
#modify it under the terms of the GNU Lesser General Public
#License as published by the Free Software Foundation; either
#version 2.1 of the License, or (at your option) any later version.

#This library is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#Lesser General Public License for more details.

#You should have received a copy of the GNU Lesser General Public
#License along with this library; if not, write to the Free Software
#Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

package ParameterTuning;
require Exporter;
@ISA = qw(Exporter);
@EXPORT_OK = qw(InitializeEnvironment RunBlackBox RunTest);
use strict;
use Data::Dumper;
use File::Slurp qw(read_file write_file);
use ARFF qw(ARFFToCSV InitializeARFF ARFFParam ARFFResult ARFFError ARFFFinish);

######
#each returns the minimum or maximum of all arguementes
sub min {
  my $min = shift;
  $min > $_ and $min = $_ foreach @_;
  return $min;
}

sub max {
  my $max = shift;
  $max < $_ and $max = $_ foreach @_;
  return $max;
}

our $black_box_cmd;
our $black_box_log_file;
our $black_box_report_file;
our @black_box_results;

#######################
#InitializeEnvironment
# sets up black box testing
# parameters are in the following order:
#   command string to run black box program
#   file name of log file (to check for errors)
#   file name of report file
#   anonymous array containing the names of each of the results
sub InitializeEnvironment($$$$) {
  $black_box_cmd = shift;
  $black_box_log_file = shift;
  $black_box_report_file = shift;
  @black_box_results = ('execution_time', @{shift()});
}

#######################
#RunTest
#allowable arguement combinations
# exhaustive test:  params => \@(param array structure)
# specific tests: params => \@(param array structure), tests => \@(array of 1-based test numbers to run (1st test = 1))
# random test: params => \@(param array structure), num_tests => $(total number of tests to perform)
# search test: params => \@(param array structure), num_tests => $(total number of tests to perform),
#               search_rate_half_life => $(test number of the half-life of the search narrowing -simulated annealing)
#               fitness_func => \&(fitness function
#                                   -receives a hash keys = result measure (@black_box_results), value = result value
#                                   -returns single number for fitness of the results, higher is better
#               simult_params => $(number of parameters to change at the same time -defaults to 2)
sub RunTest(@) {
  my %args = @_;
  #ref to hash of parameters, and array of in-order pointers to the parameters
  my ($params, $param_order) = BuildParams($args{params});

  #keys contain parameters of previous tests joined with commas,
  # values contain a reference to an array of the in-order results
  our %prev_tests;

  #compute the total number of possible tests,
  # and if the number of tests isn't specified, do all
  my $num_possible_tests = 1;
  $num_possible_tests *= @{$_->{allowed}} foreach @$param_order;
  my $num_tests = $args{num_tests} || $num_possible_tests;

  my @tests_to_run = (exists $args{tests} ? @{$args{tests}} : (1..$num_tests));

  InitializeARFF('parameter_log.arff', 'black-box', $param_order, \@black_box_results);
  foreach my $test_number (@tests_to_run) {
    print "executing test $test_number... ";

    #select the test's parameters based on how this function was called
    if(exists $args{fitness_func}) {
      SelectSearchParams($test_number-1, $num_tests, $param_order,
			 [values %prev_tests], $args{search_rate_half_life}, $args{simult_params} || 2);
    }
    elsif($num_tests >= $num_possible_tests) {
      SelectExhaustiveParams($test_number-1, $num_tests, $num_possible_tests, $param_order);
    }
    else { #random
      SelectRandomParams($param_order);
    }

    #don't re-do unnecessary tests
    # string the parameter values together to give a unique name for the test, and check if it's been logged as done
    my @param_list = map {$_->{val}} @$param_order;
    my $param_string = join ',', @param_list;
    if(exists $prev_tests{$param_string}) {
      print "repeat: skipping\n";
      next;
    }

    #record + set parameters
    ARFFParam @param_list;
    SetParams($params);

    #run the test
    my $start_time = time;
    if(!RunBlackBox($black_box_cmd, $black_box_log_file)) {
      #an error occured, so write out unknowns
      ARFFError;
      $prev_tests{$param_string} = 0;
      print "error\n";
      next;
    }

    #process results
    my @test_results = (time - $start_time, split /\s+/, read_file($black_box_report_file));
#***FOR TESTING -comment previous line, and uncomment the next two
#my @test_results = ($params->{'data/drive_parameters.dat'}{'$num_drives'}{val}, $params->{'data/layout_data.dat'}{'$left_weight'}{val});
#print "results (column,row): @test_results\n";

    ARFFResult @test_results;
    #if using a fitness function, compute and store results
    if(exists $args{fitness_func}) {
      #prepare the results to be neatly passed into the function
      my %test_result_lookup;
      @test_result_lookup{@black_box_results} = @test_results;
      #compute and store the test results
      $prev_tests{$param_string} = {
					  fitness => &{$args{fitness_func}}(\%test_result_lookup),
					  params => \@param_list,
					  };
    }
    else { #just want to record that we did the test (so it doesn't repeat it)
      $prev_tests{$param_string} = 1;
    }

    print "ok\n";
  } #foreach $test_number

  ARFFFinish();
  #make a csv for easy reading
  ARFFToCSV('parameter_log.arff', 'parameter_log.csv');
}

############
#BuildParams
# builds the parameter data structure and parameter order array based on
# the array-structured test info
# pass in: array-structured test info, returns references to constructed parameter tree and order array
sub BuildParams($) {
  my $cur_params = shift;
  my @param_order; #array of in-order pointers to the parameters
  my %params; #hash of parameters

  #build param_order and params
  # iterate over every template file
  while(@$cur_params) {
    #@$cur_params is an array with pairs of (filename, param_list)
    my ($filename, $param_list) = (splice @$cur_params, 0, 2);

    #iterate over every parameter in that file
    while(@$param_list) {
      #@$param_list is an array with pairs of (param, values)
      my ($param, $vals) = (splice @$param_list, 0, 2);

      #if the first value is a hash, then treat it as a list of requirements
      # for the variable to be modified
      my $constraints = {};
      my $is_list = 0; #if true, then the parameters are not ordered
      if(ref($vals->[0]) =~ /HASH/) {
	$constraints = shift @$vals;
	$is_list = delete $constraints->{_list_} if exists $constraints->{_list_};
      }

      my $default = '';
      #if the next one starts with a '*', pull it off and use as the default value
      if($vals->[0] =~ /^\*(.*)/) {
	  $default = $1;
	  shift @$vals;
	}
      else { #pick a random default value
	$default = $vals->[rand @$vals];
      }

      #populate the parameter
      $params{$filename}{$param} = {
				    name => $param,
				    val => $default,
				    allowed => $vals,
				    constraints => $constraints,
				    list => $is_list,
				   };
      #add the parameter to the order list so it can be printed out in order
      push @param_order, $params{$filename}{$param};
    }
  }

  return \%params, \@param_order;
}

############
#SelectRandomParams
# randomly sets each current value of each parameter in @param_list
# pass in: \@param_list
sub SelectRandomParams($) {
  #pick a random value from the allowed array for each of the parameters
  # for every item in the list passed in, randomly pick one of ->{allowed} and put it in ->{val}
  $_->{val} = $_->{allowed}[rand @{$_->{allowed}}] foreach @{(shift)};
} #sub SelectParams

############
#SelectExhaustiveParams
# sets the current value of each parameter in @param_list based on
# test number (0-based) and total number of tests
# pass in: test number, total number of tests, total possible tests, \@param_list
sub SelectExhaustiveParams($$$$) {
  my ($test_number, $num_tests, $cur_combinations, $param_list) = @_;

  #choose the value corresponding to the most significant
  # "variable width" digit (where each parameter represents a digit)
  foreach my $p (@$param_list) {
    $cur_combinations /= @{$p->{allowed}}; #next "digit" based on number of allowed values

    #get current "digit", and remove current "digit" from $test_number
    $p->{val} = $p->{allowed}[$test_number / $cur_combinations];
    $test_number %= $cur_combinations;
  }
} #sub SelectParams

#############################
#GetAverageFitness
# returns the average fitness of the specified allowed values range
# takes in $min_offset, $max_offset, $param_number, $param_order, $prev_tests
sub GetAverageFitness {
  my ($min_offset, $max_offset, $param_number, $param_order, $prev_tests) = @_;
  my $p = $param_order->[$param_number];

  my ($fitness, $count) = (0, 0);
  #for each index representing a parameter value
  foreach my $i ($min_offset..$max_offset) {
    #see if there are any previous tests that used the value at index $i as the parameter value
    my @results = grep {$_->{params}[$param_number] eq $p->{allowed}[$i]} @$prev_tests;
    #results have been gathered, so take the average
    if(@results) {
      $fitness += $_->{fitness} foreach @results;
      $count += @results;
    }
  }

  return $fitness / $count if $count > 0; #find average
  return 1e10; #make it really big so there's a good chance we'll go this way
}

############
#SelectSearchParams
# sets the current value of each parameter in @param_list based on
# test number and total number of tests
# pass in: test number, total number of tests, \@parameter list,
#  \@previous tests (array of hashes of {fitness => $value, params=> \@inorder_parameter_value_list},
#  search rate half-life, simultaneous number of parameters
sub SelectSearchParams($$$$$$) {
  my ($test_number, $num_tests, $param_order, $prev_tests, $search_rate_half_life, $simult_params) = @_;

  #use default population (random and/or specified by test set) if it's the first test,
  # so it has a point to start from
  return if $test_number == 0;

  #choose the test to branch from out of any of the tests with fitness values in the top 3
  # by sorting the results based on fitness
  #***NOTE, WE MAY WISH TO CHANGE HOW MANY OF THE TOP ARE CHOSEN
  my @sorted_tests = sort {$b->{fitness} <=> $a->{fitness}} grep {$_ != 0} @$prev_tests;
  splice @sorted_tests, 3; #get rid of all but the top 3
  #if we only have dead (errored) runs, then try a new random configuration and start over
  unless(@sorted_tests) {
    SelectRandomParams($param_order);
    return;
  }
  my @new_params = @{$sorted_tests[rand @sorted_tests]->{params}}; #randomly choose a set of parameters from @sorted_tests

  #copy over parameters from the parent test
  $param_order->[$_]{val} = $new_params[$_] foreach(0..$#new_params);

  #maximum search jump "distance" is determined by the test number
  # f = e^(-t/T)  where t is the current "time" (test number),
  # and T is to scale t such that its half-life lines up where the parameter $search_rate_half_life specifies
  my $range_fraction = exp( -$test_number / ( -$search_rate_half_life / log(.5))  );

  #randomly order parameter choices
  my @order = (0..$#$param_order);#start with 0,1,2,...
  #randomize array by shuffling
  for(my $i = @order; --$i; ) {
    my $j = int rand($i+1);
    @order[$i,$j] = @order[$j,$i] unless $i == $j;
  }

  #change as many parameters as requested
 EACH_PARAMETER:
  for(1..$simult_params) {

    #pick the next random parameter (by index)
    #***should each parameter have a particular weighting?
    last unless @order; #can't proceed without another parameter
    my $param_number = shift @order;
    my $p = $param_order->[$param_number];

    #make list of parameter names for lookup
    my %param_lookup;
    $param_lookup{$_->{name}} = $_ foreach @$param_order;

    #make sure each constraint is satisifed,
    # otherwise grab another parameter and start again
    foreach my $name (keys %{$p->{constraints}}) {
	redo EACH_PARAMETER unless $param_lookup{$name}{val} eq $p->{constraints}{$name};
    }

    #if the allowed values are not in any order at all and marked as so,
    # then choose one randomly, because gradient search doesn't do any good
    if($p->{list}) {
      $p->{val} = $p->{allowed}[rand @{$p->{allowed}}];
      next;
    }

    #find offset of current value
    my $cur_index;
    foreach(0..$#{$p->{allowed}}) {
      $cur_index = $_ and last if $p->{val} eq $p->{allowed}[$_];
    }

    #get a random offset based on the number of allowed values for the parameter
    # it is based such that at the initial time, one movement will go across up to 1/2 of the range,
    # so the expected distance is 1/4th in either direction (based on a uniformly distributed random number)
    my $max_index_distance = int($range_fraction * @{$p->{allowed}} / 2 + 1);
    my $min_index = max($cur_index - $max_index_distance, 0);
    my $max_index = min($cur_index + $max_index_distance, $#{$p->{allowed}});

    #get the new index of the number to use in the parameter
    my $new_index = $cur_index;

    #if $cur_index is at one end of allowed values, then go in the other direction
    if($min_index == $cur_index) {
      $new_index = $max_index;
    } elsif($max_index == $cur_index) {
      $new_index = $min_index;
    }
    else {#not at the edge, so choose based on fitness

      my $down_fitness = GetAverageFitness($min_index, $cur_index-1, $param_number, $param_order, $prev_tests);
      my $up_fitness = GetAverageFitness($cur_index+1, $max_index, $param_number, $param_order, $prev_tests);

      my $max_prob = 0.0; #probability to go with $max_index
      if($down_fitness > $up_fitness) {
	$max_prob = 1/3.0;
      } elsif($up_fitness > $down_fitness) {
	$max_prob = 2/3.0;
      }
      else { #same
	$max_prob = 0.5;
      }

      #choose max or min index based on probability
      if(rand > $max_prob) {
	$new_index = $max_index;
      }
      else {
	$new_index = $min_index;
      }
    } #else choose min or max index

    #choose the value randomly between the old and new index
    my $r = rand;
    $new_index = int($r * $new_index + (1-$r) * $cur_index);
    $p->{val} = $p->{allowed}[$new_index];
  } #for 1..$simult_params
} #sub SelectSearchParams

############
#SetParams
# sets the parameters of the input files
# pass in a reference to a 2-deep hash of the following form:
# key: filename.ext
#  value: hash of:
#       key: parameter_name (usually preceeded by rarely used character as $)
#        value: value to replace it with
sub SetParams($) {
  my $h = shift;
  foreach my $fname (keys %$h) {

    #see if swapping files
    if($fname eq 'files') {
      #move each corresponding file over
      `cp $_->{val} $_->{name}` foreach values %{$h->{files}};
      next;
    }

    #read file, replace all occurances of each parameter, write back out
    my $template = read_file("$fname.tmpl");
    #find all parameters, and sort by the length of the name, longest first
    # this will prevent problems where one variable name contains the other (e.g. var and variable)
    foreach my $param (sort {length($b->{name}) <=> length($a->{name})}
		       values %{$h->{$fname}}) {
      my $backslashed = quotemeta $param->{name}; #backslashify any characters required
      $template =~ s/$backslashed/$param->{val}/g;
    }
    write_file($fname, $template);
  } #foreach $fname
} #sub setparams


############
#RunBlackBox
# 2 parameters: the program name to execute, the log file written by the program
# returns true if the program executed successfully, false if it didn't
sub RunBlackBox {
  my ($prog, $logfile) = @_;

  #remove logfile so we don't use an old one in case of error
  unlink $logfile;

  system($black_box_cmd);

  #return true only if the logfile exists and doesn't contain 'error'
  return -f $logfile  &&  read_file($logfile) !~ /error/i;
}

1;
