#!/usr/bin/perl
#These two lines are required
use strict;
use ParameterTuning qw(RunTest);

#fetch all of the files that end in .xyz
# export the layouts with some extension (or some pattern to easily match)
# and then use all of them as possibilities
#these two lines can obviously be copied with different variables
my @xyz_sets = `ls data/*.xyz`;
chomp @xyz_sets;  #remove extra newlines

#the main parameter block
# it is set up as follows
# 'filename' => [list of parameters],
#
# there is a special filename 'files', where the parameters are the names
#      of the files to be replaced, with the allowed values being the files to replace them
#      this substitution happens first, before any of the other variables are replaced
#      other than this, the 'files' key is treated exactly the same as regular variables
#
#
# each list of parameters is defined as follows:
#  '$variable_name' => [allowed value list],
#
#  the parameter name is exactly what will appear in the excel spreadsheet
#   the files that contain variables should be copied and appended with .tmpl for "template"
#   to make the variables distinct, a good convention is to prefix them with dollar signs $
#
# each allowed value list is an array of values that the parameter may take
#  two special values are allowed first
#  the first in this list may optionally be a hash of attributes of the parameter
#    if the value of _list_ is 1 in the hash, then the parameter is treated as a list of
#      values, of which will not be applicable to gradient search (it will be randomly chosen)
#      this is useful for cases when no logical relationship exists between values,
#      such as trying out different layouts
#    the other values in this hash are treated as requirements for the parameter to be active
#      for example, if it contains  '$some_var' => 'TRUE'  then the parameter will only be modified
#      when $some_var is set to 'TRUE'.  this is helpful in constraining the search space
#      so that it does not get stuck on conditionally unused parameters, such as when an algorithm
#      may be turned on or off in addition to its parameters changed.
#  optionally following the optional attribute hash is the initial value of the parameter
#    this is indicated by prefixing it with an asterisk, such as '*5'.  if this parameter is not
#    specified, one will be randomly chosen
#  following those two optional parameters, is the array of allowed values for the parameter
#    any valid perl expression is allowed, such as:
#    1,2,7,10,0.8,3.5,'TRUE',7,'ON'   #list of values
#    0..10,12..15                     #range of 0 through 10 and 12 through 15 (0-15 excluding 11)
#    map {$_/100} 0..100              #0.0, 0.01, 0.02, ... 0.98,0.99,1.0
#
#  thus, the following are all valid allowed value lists:
#  '$param' => [ 'ON', 'OFF' ],
#  '$param' => [ '*ON', 'ON', 'OFF'],
#  '$param' => [ {_list_ => 1}, '*algorithm1', 'algorithm1', 'algorithm2', 'algorithm3'],   #nominal data (algorithms are not in any particular order)
#  '$param' => [ {'$param2' => 1, _list => 1}, 'a', 'b', 'c'],	# $param2 must be 1 to search on this value (otherwise it does nothing)
#  '$param' => [ '*0', map {$_/5} 0..5],	#values 0.0, 0.2, 0.4, 0.6, 0.8, 1.0, with 0 being default
#  '$param' => [ {'$param2' => 'ON'}, 0, 0.5, 1, 2],

my $params = [
  'files' => [
	      'xyz.dat' => [
					    {_list_ => 1},
					    @xyz_sets],
  ],

  'data1.dat' => [
			   '$algorithm1' => ['on', 'off'],
			   '$alg_param1' => [
                                            {'$algorithm1' => 'on'},
					    1..10],
  ],

  ],
  'data2.dat' => [
			   '$var1' => ['*8', 1..20],
  ],
];


InitializeEnvironment('./blackbox.pl > ./out.txt', './out.txt', 'output.dat', ['r1', 'r2']);

########
#the fitness function is used with gradient search
# it recieves a reference of a hash with keys being the result names,
# and values being one of the initialized results
# which is one of:
# execution_time orders_filled lines_completed items_replen jobs_completed recircs time_per_pod order_fill_time du_util du_energy du_dist du_congestion operator_util bin_utilization
# the function must return a numeric value indicating the fitness of the simulation
# the only relationship required is that higher numbers mean "better", but besides that,
# they can have any form of relationship (linear, logarithmic), and can be any size or range
sub fitness {
  my $args = shift;
  return $args->{r1};
}

######
#running the test
#
# run all combinations of parameters exhaustively:
#  RunTest(params => $params);
#  obviously you don't want to do this unless you only have a couple of parameters with few values
#
# run a 50 random combinations of the parameters' values
#  RunTest(params => $params, num_tests => 50);
#
# run a set of specified tests, where the test numbers are the index into the configuration space
#  useful for design of experiment techniques
#  e.g. parameter A having values 'FALSE', 'TRUE', parameter B having values 1,2,3
#    test number   A     B
#         1       FALSE  1
#         2       FALSE  2
#         3       FALSE  3
#         4       TRUE   1
#         5       TRUE   2
#         6       TRUE   3
#   RunTest(params => $params, tests => [1,5,4,6]);
#
# perform a stochastic gradient search on 100 tests,
#  using the fitness function to evaluate the results, and will narrow in on the
#  results with a half-life at the 40th test.  at the first test, each parameter
#  may be permuted up to 1/2 the total range of the parameters (1/4 in each direction).
#  the half-life will be the point at which it will be maximally permuted 1/4 the total range.
#  simult_params means that 4 parameters will be changed at a time.  increasing this number
#  will result in faster searching, but has the possibility of overshooting maxima
# RunTest(params => $params, num_tests => 100, fitness_func => \&fitness,
#	search_rate_half_life => 40, simult_params => 4);

RunTest(params => $params, num_tests => 100, fitness_func => \&fitness,
	search_rate_half_life => 50, simult_params => 4);
